#include <windows.h>
#include <iostream>
#include <sstream>

#include "aarot.hpp"

using namespace std;

#define IMAGENAME "test.bmp"
#define DEGREES 64.1

HBITMAP hrotbmp;
BITMAP bitmapinfo;

bool callbackfunc(double percentdone)
{
	ostringstream out;
	out << percentdone * 100;
	if (out.str().length() <= 5)
		cout << out.str() << "%\tdone" << endl;
	else
		cout << out.str().substr(0, 5) << "%\tdone" << endl;
    return false;
}

LRESULT CALLBACK WindowProcedure (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_PAINT:
            {
                HDC drawdc = GetDC(hwnd);
                HDC srcdc = CreateCompatibleDC(0);
                
                HBITMAP holdbmp = (HBITMAP)SelectObject(srcdc, hrotbmp);

                BitBlt(drawdc, 0, 0, bitmapinfo.bmWidth, bitmapinfo.bmHeight, srcdc, 0, 0, SRCCOPY);

                SelectObject(srcdc, holdbmp);
                DeleteDC(srcdc);
                
                ReleaseDC(hwnd, drawdc);
            }
            break;
        case WM_DESTROY:
            PostQuitMessage (0);
            break;
    }
    return DefWindowProc (hwnd, message, wParam, lParam);
}

int main()
{
	return WinMain(NULL, NULL, NULL, SW_SHOW);
}

int WINAPI WinMain (HINSTANCE hThisInstance, HINSTANCE hPrevInstance, LPSTR lpszArgument, int nFunsterStil)
{
    //Load and rotate image
    HBITMAP hsrcbmp = (HBITMAP)LoadImage(NULL, IMAGENAME, IMAGE_BITMAP, 0, 0, LR_DEFAULTSIZE | LR_LOADFROMFILE);

    int starttime = GetTickCount();
    hrotbmp = aarot::rotate(hsrcbmp, DEGREES, callbackfunc, 0xFFFFFF, true);
    cout << "Rotation completed in " << GetTickCount() - starttime << " milliseconds" << endl;    

    DeleteObject(hsrcbmp);

    //Get info on the image so we can create the right sized window and for drawing
    GetObject(hrotbmp, sizeof(BITMAP), &bitmapinfo);

    //Standard windows code
    HWND hwnd;
    MSG messages;
    WNDCLASSEX wincl;

    wincl.hInstance = hThisInstance;
    wincl.lpszClassName = "Rotator";
    wincl.lpfnWndProc = WindowProcedure;
    wincl.style = CS_DBLCLKS;
    wincl.cbSize = sizeof (WNDCLASSEX);
    wincl.hIcon = LoadIcon (NULL, IDI_APPLICATION);
    wincl.hIconSm = LoadIcon (NULL, IDI_APPLICATION);
    wincl.hCursor = LoadCursor (NULL, IDC_ARROW);
    wincl.lpszMenuName = NULL;
    wincl.cbClsExtra = 0;
    wincl.cbWndExtra = 0;
    wincl.hbrBackground = (HBRUSH) COLOR_BACKGROUND;

    if (!RegisterClassEx (&wincl))
        return 0;

    hwnd = CreateWindowEx (0, "Rotator", "Rotation Example", WS_OVERLAPPEDWINDOW,  CW_USEDEFAULT,  CW_USEDEFAULT, 
           bitmapinfo.bmWidth + 8, bitmapinfo.bmHeight + 34, HWND_DESKTOP, NULL, hThisInstance, NULL);    

    ShowWindow (hwnd, nFunsterStil);
    while (GetMessage (&messages, NULL, 0, 0))
    {
        TranslateMessage(&messages);
        DispatchMessage(&messages);
    }
    return (int)messages.wParam;
}
